from .path import Path
from .base import ObjectOnFileSystem
from .file import File
from .utils import outputs_observable
import rx


class Directory(ObjectOnFileSystem):
    def __init__(self, path: Path, filesystem=None):
        super().__init__(filesystem, path)

    def exists(self):
        result = super().exists()
        if not result:
            return result
        with self.filesystem.open() as fs:
            if not fs.isdir(self.path.s):
                raise NotADirectoryError(self.path.s)
        return result

    def listdir(self):
        with self.filesystem.open() as fs:
            children = fs.listdir(self.path.s)
            paths = [self.path / c for c in children]
            results = []
            for p in paths:
                if fs.isfile(p.s):
                    results.append(File(p, self.filesystem))
                else:
                    results.append(Directory(p, self.filesystem))
            return tuple(results)

    def listdir_as_observable(self):
        return rx.Observable.from_(self.listdir(),
                                   scheduler=rx.concurrency.ThreadPoolScheduler())

    def copy_to(self, target_path: Path):
        with self.filesystem.open() as fs:
            fs.copydir(self.path.s, target_path.s)
        return Directory(target_path, self.filesystem)

    def sync(self, source: 'Observable[ObjectOnFileSystem] or List[ObjectOnFileSystem]'):
        is_observable = isinstance(source, rx.Observable)
        if not isinstance(source, rx.Observable):
            source = rx.Observable.from_(source,
                                         scheduler=rx.concurrency.ThreadPoolScheduler())
        with self.filesystem.open() as fs:
            result = (source.map(lambda o: o.copy_to(self.path / o.path.n))
                      .to_list()
                      .to_blocking()
                      .first())
        if is_observable:
            return rx.Observable.from_(result,
                                       scheduler=rx.concurrency.ThreadPoolScheduler())
        return result

    def attach_file(self, name: str) -> File:
        """
        Parameters:

        - `filename`: 

        """
        return File(self.path / name, self.filesystem)

    def attach_directory(self, name: str) -> 'Directory':
        return Directory(self.path / name, self.filesystem)

    def remove(self):
        with self.filesystem.open() as fs:
            fs.removetree(self.path.s)

    def makedir(self, name: str) -> 'Directory':
        with self.filesystem.open() as fs:
            raw_path = (self.path / name).s
            fs.makedir(raw_path)
            return Directory(raw_path, self.filesystem)


def match_directory(patterns):
    if isinstance(patterns, str):
        patterns = (patterns, )
    return lambda o: isinstance(o, Directory) and (o.match(patterns))


def match_file(patterns):
    if isinstance(patterns, str):
        patterns = (patterns,)
    from .file import File
    return lambda o: isinstance(o, File) and (o.match(patterns))



from fs import path as fp
from typing import Tuple


class Path:
    """ Unified path for posix/windows/url/url_quoted paths.
    For the purpose of unified process, %,: and @ are preserved, please make sure there is not such characters in raw path.
    This class is an 'abstract' path object, which means it only normalize its representation for different platform,
    and provide functionalities of pure path calculations, without any relation to the true file system.
    Thus this class does **NOT** provide methods like `exists()` or `is_dir()`, please refer `dxpy.file_syste.file.File` for these functions.
    """

    @classmethod
    def _divide_protocol(cls, p: str) -> (str, str):
        n = p.find('@')
        if n >= len(p) - 1:
            return p[:], ''
        if n == -1:
            return None, p
        else:
            return p[:n + 1], p[n + 1:]

    @classmethod
    def _decode_url(self, p: str):
        from urllib.parse import unquote_plus
        return unquote_plus(unquote_plus(p))

    @classmethod
    def _unified_path(cls, p) -> str:
        if isinstance(p, Path):
            return p._p
        if not isinstance(p, str):
            raise TypeError(
                "{} is not convertable to {}.".format(type(p), __class__))
        protocol, raw_path = cls._divide_protocol(p)
        raw_path = cls._decode_url(raw_path)
        raw_path = fp.normpath(raw_path)
        if protocol is None:
            return raw_path
        return protocol + raw_path

    @classmethod
    def _unified_protocol(cls, p: str) -> str:
        if not p.endswith('@'):
            return p + '@'
        return p

    def __init__(self, path, protocol=None):
        if protocol is not None:
            path = self._unified_protocol(protocol) + path
        self._p = self._unified_path(path)

    @property
    def protocol(self) -> str:
        return self._divide_protocol(self._p)[0]

    @property
    def raw(self) -> str:
        return self._divide_protocol(self._p)[1]

    def raw_path(self) -> 'Path':
        """
        Returns raw path object, i.e. no protocol.
        """
        return Path(self.raw, None)

    @property
    def s(self) -> str:
        "Alias of self.raw_path"
        return self.raw

    @property
    def a(self) -> str:
        """
        Returns absolute path in str.
        """
        return fp.abspath(self.s)

    def absolute(self) -> 'Path':
        return Path(self.a, self.protocol)

    @property
    def r(self) -> str:
        return fp.relpath(self.s)

    def relative(self) -> 'Path':
        return Path(self.r, self.protocol)

    def is_root(self) -> bool:
        return self.s == '/'

    def is_absolute(self) -> bool:
        return fp.isabs(self.s)

    def is_relative(self) -> bool:
        return not self.is_absolute()

    @property
    def n(self) -> str:
        return fp.basename(self.s)

    def name(self) -> 'Path':
        return Path(self.n, self.protocol)

    @property
    def extension_name(self) -> str:
        return fp.splitext(self.s)[1]

    @property
    def e(self) -> str:
        return self.extension_name

    @property
    def f(self) -> str:
        return fp.dirname(self.s)

    def father(self) -> 'Path':
        return Path(self.f, self.protocol)

    def parts(self) -> Tuple[str]:
        result = list(fp.iteratepath(self.s))
        if self.is_absolute():
            result = ['/'] + result
        return tuple(result)

    def join(self, name):
        return Path(fp.combine(self.s, name), self.protocol)

    def __truediv__(self, name):
        return self.join(name)

    def __add__(self, suffix):
        return Path(self.s + suffix, self.protocol)

    def __str__(self):
        return self._p

    def __hash__(self):
        return hash(self._p)

    def __eq__(self, path):
        return self._p == Path(path)._p

    def regex_match(self, pattern: str, is_ignore_extension=False) -> bool:
        raise NotImplementedError

    def unix_match(self, pattern: str, is_ignore_extension=False) -> bool:
        raise NotImplementedError

    def url_safe_path(self, with_protocol=False) -> str:
        from urllib.parse import quote_plus
        result = quote_plus(quote_plus(self.s))
        if with_protocol:
            return self.protocol + result
        return result



from .base import ObjectOnFileSystem, FileSystem
from .path import Path
import fs
from .conf import mc


class NotAFileError(Exception):
    def __init__(self, path: str):
        super().__init__("{} is not a file.".format(path))


class File(ObjectOnFileSystem):
    """
    Representation of File.
    """

    def __init__(self, path: Path, filesystem: FileSystem=None):
        super().__init__(filesystem, path)

    def name(self):
        return self.path.n

    def suffix(self):
        return self.path.suffix

    def load(self) -> 'File':
        """
        Returns:

        - contents loaded by assuming file as binary str.
        """
        if not self.exists():
            raise FileNotFoundError(self.path.abs)
        with self.filesystem.open() as fs:
            with fs.open(self.path.s, 'rb') as fin:
                return fin.read()

    def save(self, data: 'str or bytes'):
        if isinstance(data, str):
            data = data.encode()
        with self.filesystem.open() as fs:
            with fs.open(self.path.s, 'wb') as fout:
                return fout.write(data)

    def exists(self):
        result = super().exists()
        if not result:
            return result
        with self.filesystem.open() as fs:
            if not fs.isfile(self.path.s):
                raise NotAFileError(self.path.s)
        return result

    def copy_to(self, target_path: Path):
        with self.filesystem.open() as fs:
            fs.copy(self.path.s, target_path.s, True)
        return File(target_path, self.filesystem)

    def remove(self):
        with self.filesystem.open() as fs:
            fs.remove(self.path.s)



from functools import wraps
import rx


def outputs_observable(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        result = func(*args, **kwargs)
        if isinstance(result, rx.Observable):
            return result
        return rx.Observable.from_(tuple(result))
    return wrapper



from setuptools import setup, find_packages
setup(name='dxl-fs',
      version='0.1.5',
      description='File system library.',
      url='https://github.com/Hong-Xiang/dxl',
      author='Hong Xiang',
      author_email='hx.hongxiang@gmail.com',
      license='MIT',
      namespace_packages=['dxl'],
      packages=find_packages('src/python'), 
      package_dir = {'': 'src/python'},
      install_requires=[
          'click',
          'rx',
          'fs',
      ],
      scripts=['src/cli/dxfs.py'],
    #   namespace_packages = ['dxl'],
      zip_safe=False)



from contextlib import contextmanager
from typing import Callable, TypeVar

import fs.base
import fs.osfs

from .conf import mc
from .path import Path


class FileSystem:
    """
    Abstract filesystem. Could be one of these states:

    - Instance of fs.base.FS
    - Callable[[], Instance of fs.base.FS], including fs.base.FS itself (like OSFS).

    Note FileSystem is designed to work in lazy, thus before `with self.open()` is called,

    """

    def __init__(self, filesystem=None, base_path='.'):
        if isinstance(filesystem, FileSystem):
            self.filesystem = filesystem.filesystem
            self.base_path = filesystem.base_path
        else:
            self.filesystem = filesystem
            self.base_path = base_path

    def _fs_args(self, base_path=None):
        import fs.memoryfs
        if self.filesystem == fs.memoryfs.MemoryFS:
            return ()
        else:
            if base_path is None:
                return (self.base_path,)
            else:
                return (base_path,)

    @contextmanager
    def open(self, base_path=None) -> fs.base.FS:
        """
        Returns opened instance of fs.base.FS.
        """
        try:
            if isinstance(self.filesystem, fs.base.FS):
                self.filesystem.check()
                if base_path is not None:
                    raise TypeError(
                        "Argument base_path must be non for instanced filesystem.")
                yield self.filesystem
            else:
                with self.filesystem(*self._fs_args(base_path)) as fs_instance:
                    yield fs_instance
        except Exception as e:
            raise e

    @property
    def info(self):
        """
        Returns Dict of basic types or str, i.e. able to serialized by JSON.
        """
        with self.open() as fs:
            return str(fs)


class ObjectOnFileSystem:
    def __init__(self, filesystem: FileSystem, path: Path):
        if filesystem is None:
            filesystem = mc.default_filesystem
        self.filesystem = FileSystem(filesystem)
        self.path = Path(path)

    @property
    def info(self):
        return {'filesystem': self.filesystem.info,
                'path': self.path.s}

    def exists(self):
        with self.filesystem.open() as fs:
            return fs.exists(self.path.s)

    def match(self, patterns):
        with self.filesystem.open() as sfs:
            return sfs.match(patterns, self.path.s)

    def system_path(self):
        import fs.errors
        with self.filesystem.open() as sfs:
            try:
                return Path(sfs.getsyspath(self.path.s)).s
            except fs.errors.NoSysPath:
                return self.path.s

    def copy_to(self, target_path):
        raise NotImplementedError

    def remove(self):
        raise NotImplementedError
